from typing import Any, Dict, List, Tuple, Optional, Union

from dataclasses import dataclass
from datetime import datetime

import re
import sys, tempfile, os
from subprocess import run
import yaml

import logging
import rich
from rich.console import Console
from rich.table import Table

from terka.domain.task import Task
from terka.domain.project import Project
from terka.domain.user import User
from terka.domain.commentary import TaskCommentary, ProjectCommentary, EpicCommentary, StoryCommentary, SprintCommentary
from terka.domain.tag import BaseTag, TaskTag, ProjectTag
from terka.domain.collaborators import TaskCollaborator, ProjectCollaborator
from terka.domain.event_history import TaskEvent, ProjectEvent
from terka.domain.sprint import Sprint, SprintTask
from terka.domain.time_tracker import TimeTrackerEntry
from terka.domain.epic import Epic, EpicTask
from terka.domain.story import Story, StoryTask
from terka.domain.external_connectors.asana import AsanaTask, AsanaProject

from terka.service_layer import services, printer
from terka.service_layer.ui import TerkaTask
from terka.adapters.repository import AbsRepository
from terka.utils import format_command, format_task_dict

logger = logging.getLogger(__name__)


@dataclass
class CurrentEntry:
    value: str
    type: str
    source: str = None


def generate_message_template(current_entry: CurrentEntry, task: Task,
                              repo) -> str:
    if isinstance(task, (Task, Story, Epic)):
        project = services.lookup_project_name(task.project, repo)
        project_name = project.name
    elif isinstance(task, Project):
        project_name = task.name
    if current_entry.source == "sprints":
        message_template = f"""
        # You are editing {current_entry.type}, enter below:
        {current_entry.value}
        ---
        {current_entry.source} context:
        id: {task.id}
        goal: {task.goal}
        """
    else:
        message_template = f"""
        # You are editing {current_entry.type}, enter below:
        {current_entry.value}
        ---
        {current_entry.source} context:
        id: {task.id}
        name: {task.name}
        description: {task.description}
        project: {project_name}
        """
    return re.sub("\n +", "\n", message_template.lstrip())


class BaseHandler:

    def __init__(self, successor) -> None:
        self._successor = successor

    def handle(self, entity):
        if self._successor:
            return self._successor.handle(entity)
        return None


class TaskHandler(BaseHandler):

    def handle(self, entity):
        if "tasks".startswith(entity):
            logger.debug("Handling task")
            return Task, "tasks"
        return super().handle(entity)


class ProjectHandler(BaseHandler):

    def handle(self, entity):
        if "projects".startswith(entity):
            logger.debug("Handling project")
            return Project, "projects"
        return super().handle(entity)


class UserHandler(BaseHandler):

    def handle(self, entity):
        if "users".startswith(entity):
            logger.debug("Handling user")
            return User, "users"
        return super().handle(entity)


class TagHandler(BaseHandler):

    def handle(self, entity):
        if "tags".startswith(entity):
            logger.debug("Handling tags")
            return BaseTag, "tags"
        return super().handle(entity)


class SprintHandler(BaseHandler):

    def handle(self, entity):
        if "sprints".startswith(entity):
            logger.debug("Handling sprints")
            return Sprint, "sprints"
        return super().handle(entity)


class SprintTaskHandler(BaseHandler):

    def handle(self, entity):
        if entity == "sprint_tasks":
            logger.debug("Handling sprint task")
            return SprintTask, "sprint_tasks"
        return super().handle(entity)


class TimeTrackerHandler(BaseHandler):

    def handle(self, entity):
        if entity == "time_entry":
            logger.debug("Handling time entry")
            return TimeTrackerEntry, "time_entries"
        return super().handle(entity)


class EpicHandler(BaseHandler):

    def handle(self, entity):
        if "epics".startswith(entity):
            logger.debug("Handling epics")
            return Epic, "epics"
        return super().handle(entity)


class EpicTaskHandler(BaseHandler):

    def handle(self, entity):
        if entity == "epic_tasks":
            logger.debug("Handling epic task")
            return EpicTask, "epic_tasks"
        return super().handle(entity)


class StoryHandler(BaseHandler):

    def handle(self, entity):
        if entity in ("story", "stories"):
            logger.debug("Handling stories")
            return Story, "stories"
        return super().handle(entity)


class StoryTaskHandler(BaseHandler):

    def handle(self, entity):
        if entity == "story_tasks":
            logger.debug("Handling epic task")
            return StoryTask, "story_tasks"
        return super().handle(entity)


class CommandHandler:

    def __init__(self, repo: AbsRepository):
        self.repo = repo
        self.handler = self._init_handlers()
        self.home_dir = os.path.expanduser('~')
        self.printer = printer.Printer(repo)
        self.console = Console()
        self.config = self._read_config()

    def _init_handlers(self):
        handler_chain = BaseHandler(None)
        for handler in [
                TaskHandler, ProjectHandler, UserHandler, TagHandler,
                SprintHandler, SprintTaskHandler, TimeTrackerHandler,
                EpicHandler, StoryHandler
        ]:
            new_handler = handler(handler_chain)
            handler_chain = new_handler
        return handler_chain

    def handle(self, entity):
        return self.handler.handle(entity)

    def execute(self,
                command,
                entity_type=None,
                kwargs: Dict[str, Any] = None):
        session = self.repo.session
        if entity_type:
            entity, entity_type = self.handle(entity_type)
        else:
            entity = None
        if not entity and command not in ("init", "unfocus", "log", "calendar",
                                          "help"):
            raise ValueError(f"Entity *{entity_type}* is not a valid entity")
        command = format_command(command)
        if command == "list":
            print_options = printer.PrintOptions(
                show_tasks=False,
                show_history=bool(kwargs.get("show_history")),
                show_commentaries=bool(kwargs.get("show_commentaries")),
                show_completed=bool(kwargs.get("show_completed")))
            if entity_type == "tasks":
                if "status" not in kwargs and "all" not in kwargs:
                    kwargs["status"] = "BACKLOG,TODO,IN_PROGRESS,REVIEW"
                else:
                    if "all" in kwargs:
                        del kwargs["all"]
                    show_completed = True
            if entity_type in ("stories", "epics"):
                if "all" in kwargs:
                    kwargs["status"] = "ACTIVE,COMPLETED"
                    del kwargs["all"]
                    print_options.show_completed = True
                else:
                    kwargs["status"] = "ACTIVE"
            if entity_type == "sprints":
                if "all" in kwargs:
                    kwargs["status"] = "PLANNED,ACTIVE,COMPLETED"
                    del kwargs["all"]
                else:
                    kwargs["status"] = "PLANNED,ACTIVE"
            if entity_type == "projects":
                if "all" in kwargs:
                    kwargs["status"] = "DELETED,ACTIVE,ON_HOLD,COMPLETED"
                    del kwargs["all"]
                    show_completed = True
                else:
                    kwargs["status"] = "ACTIVE"

            if (custom_sort := kwargs.get("sort")):
                del kwargs["sort"]
            else:
                custom_sort = None
            if entity_type == "projects" and "project_id" in kwargs:
                kwargs["id"] = kwargs.pop("project_id")
            if has_collaborators := "collaborators" in kwargs:
                collaborator_name = kwargs["collaborators"]
                collaborator = self.repo.list(User,
                                              {"name": collaborator_name})
                if collaborator:
                    task_collaborators = self.repo.list(
                        TaskCollaborator, {"collaborator": collaborator[0].id})
                    if task_collaborators:
                        tasks_with_collaborators = set([
                            task_collaborator.task
                            for task_collaborator in task_collaborators
                        ])
                        del kwargs["collaborators"]
                    else:
                        self.console.print(
                            f"[red]No task with collaborator '{colaborator_name}' found![/red]"
                        )
                        exit()
                else:
                    self.console.print(
                        f"[red]No user '{collaborator_name}' found![/red]")
                    exit()
            else:
                tasks_with_collaborators = None
            if has_tags := "tags" in kwargs:
                tag_text = kwargs["tags"]
                tag = self.repo.list(BaseTag, {"text": tag_text})
                if tag:
                    task_tags = self.repo.list(TaskTag, {"tag": tag[0].id})
                    if task_tags:
                        tasks_with_tag = set(
                            [task_tag.task for task_tag in task_tags])
                        del kwargs["tags"]
                    else:
                        self.console.print(
                            f"[red]No tasks with tag '{tag_text}' found![/red]"
                        )
                        exit()
                else:
                    self.console.print(
                        f"[red]No tag '{tag_text}' found![/red]")
                    exit()
            else:
                tasks_with_tag = None
            if tasks_with_tag and tasks_with_collaborators:
                filtered_tasks = tasks_with_tag.intersection(
                    tasks_with_collaborators)
            elif tasks_with_tag:
                filtered_tasks = tasks_with_tag
            elif tasks_with_collaborators:
                filtered_tasks = tasks_with_collaborators
            else:
                filtered_tasks = None
            if filtered_tasks:
                kwargs["id"] = ",".join(
                    [str(task_id) for task_id in list(filtered_tasks)])
            elif has_tags or has_collaborators:
                exit("No tasks found")
            entities = self.repo.list(entity, kwargs)
            if custom_sort == "due_date":
                entities_with_due_date = []
                entities_without_due_date = []
                for entity in entities:
                    if entity.due_date:
                        entities_with_due_date.append(entity)
                entities_with_due_date.sort(key=lambda c: c.due_date,
                                            reverse=False)
                entities = entities_with_due_date
            session.commit()
            self.printer.print_entities(entities,
                                        entity_type,
                                        self.repo,
                                        custom_sort,
                                        print_options=print_options)
            logger.info("<list> %s", entity_type)
            return entities, None, None
        elif command == "help":
            print("""
            available commands: 'list', 'show', 'create', 'update', 'done', 'calendar', 'log', 'edit'
            available entities: 'tasks', 'projects', 'commentaries'
            """)
        elif command == "init":
            home_dir = os.path.expanduser('~')
            path = os.path.join(home_dir, ".terka")
            if not os.path.exists(path):
                answer = input(
                    f"Do you want to init terka in this directory {path}? [Y/n]"
                )
                if "y" in answer.lower():
                    os.mkdirs(path)
                    with open(os.path.join(path, "config.yaml"), "w") as f:
                        yaml.dump({"user": "admin"}, f)
                elif "n" in answer.lower():
                    path = input("Specify full path to the terka directory: ")
                    os.mkdirs(path)
                else:
                    exit()
            elif not os.path.exists(os.path.join(path, "config.yaml")):
                answer = input(
                    f"Config.yaml not found in {path}, Create it? [Y/n]")
                if "y" in answer.lower():
                    with open(os.path.join(path, "config.yaml"), "w") as f:
                        yaml.dump({"user": "admin"}, f)
                else:
                    exit()
            else:
                print("Terka directory already exist.")
        elif command == "get":
            if entity_type == "tasks" and "status" not in kwargs:
                kwargs["status"] = "BACKLOG,TODO,IN_PROGRESS,REVIEW,DONE"
            if (custom_sort := kwargs.get("sort")):
                del kwargs["sort"]
            else:
                custom_sort = None
            if entity_type == "projects" and "project_id" in kwargs:
                kwargs["id"] = kwargs.pop("project_id")
            entities = self.repo.list(entity, kwargs)
            if custom_sort == "due_date":
                entities_with_due_date = []
                entities_without_due_date = []
                for entity in entities:
                    if entity.due_date:
                        entities_with_due_date.append(entity)
                entities_with_due_date.sort(key=lambda c: c.due_date,
                                            reverse=False)
                entities = entities_with_due_date
            session.commit()
            return entities
        elif command == "calendar":
            kwargs["sort"] = "due_date"
            self.execute("list", "tasks", kwargs)
        elif command == "log":
            table = Table(box=rich.box.SIMPLE)
            with open(f"{self.home_dir}/.terka/terka.log", "r") as f:
                head = f.readlines()
            for column in ("date", "source", "level", "message"):
                table.add_column(column)
            num_log_entries = int(kwargs.get("num_log_entries", 10))
            tail = head[-num_log_entries:]
            for row in tail[::-1]:
                info, message = row.split("] ")
                date, source, level = info.split("][")
                table.add_row(re.sub("\[", "", date), source, level, message)
            self.console.print(table)
        elif command == "focus":
            if entity_type == "tasks":
                self.config["task_id"] = kwargs["id"]
                if "project_name" in self.config.keys():
                    del self.config["project_name"]
            if entity_type == "projects":
                self.config["project_name"] = kwargs["id"]
                if "task_id" in self.config.keys():
                    del self.config["task_id"]
            with open(f"{self.home_dir}/.terka/config.yaml",
                      "w",
                      encoding="utf-8") as f:
                yaml.dump(self.config, f)
            logger.info("<focus> %s: %s", entity_type, "")
        elif command == "unfocus":
            if "task_id" in self.config.keys():
                del self.config["task_id"]
            if "project_name" in self.config.keys():
                del self.config["project_name"]
            with open(f"{self.home_dir}/.terka/config.yaml",
                      "w",
                      encoding="utf-8") as f:
                yaml.dump(self.config, f)
            logger.info("<unfocus> %s: %s", entity_type, "")
        elif command == "count":
            entities = self.repo.list(entity, kwargs)
            print(len(entities))
            logger.info("<count> tasks")
        elif command == "collaborate":
            if entity_type == "tasks":
                task_id = kwargs["id"]
                existing_task = self.repo.list(Task, {"id": task_id})
                if not existing_task:
                    raise ValueError(f"Task with id {task_id} is not found!")
                user = self.repo.list(User, {"name": kwargs["name"]})
                if not user:
                    user = User(**kwargs)
                    self.repo.add(user)
                    session.commit()
                    user_id = user.id
                else:
                    user_id = user[0].id
                existing_task_collaborator = self.repo.list(
                    TaskCollaborator, {
                        "task": task_id,
                        "collaborator": user_id
                    })
                if not existing_task_collaborator:
                    task_ids = get_ids(task_id)
                    for task_id in task_ids:
                        obj = TaskCollaborator(id=task_id,
                                               collaborator_id=user_id)
                        self.repo.add(obj)
                    session.commit()
            elif entity_type == "projects":
                project_id = kwargs["id"]
                existing_project = self.repo.list(Project, {"id": project_id})
                if not existing_project:
                    raise ValueError(
                        f"Project with id {project_id} is not found!")
                user = self.repo.list(User, {"name": kwargs["name"]})
                if not user:
                    user = User(**kwargs)
                    self.repo.add(user)
                    session.commit()
                    user_id = user.id
                else:
                    user_id = user[0].id
                existing_project_collaborator = self.repo.list(
                    ProjectCollaborator, {
                        "project": project_id,
                        "collaborator": user_id
                    })
                if not existing_project_collaborator:
                    project_ids = get_ids(project_id)
                    for project_id in project_ids:
                        obj = ProjectCollaborator(id=project_id,
                                                  collaborator_id=user_id)
                        self.repo.add(obj)
                        session.commit()
        elif command == "tag":
            if entity_type == "tasks":
                task_id = kwargs["id"]
                existing_task = self.repo.list(Task, {"id": task_id})
                if not existing_task:
                    raise ValueError(f"Task with id {task_id} is not found!")
                tag_text = kwargs.get("text") or kwargs.get("tags")
                tag_info = {"text": tag_text}
                tag = self.repo.list(BaseTag, tag_info)
                if not tag:
                    tag = BaseTag(**tag_info)
                    self.repo.add(tag)
                    session.commit()
                    tag_id = tag.id
                else:
                    tag_id = tag[0].id
                existing_task_tag = self.repo.list(TaskTag, {
                    "task": task_id,
                    "tag": tag_id
                })
                if not existing_task_tag:
                    task_ids = get_ids(task_id)
                    for task_id in task_ids:
                        obj = TaskTag(id=task_id, tag_id=tag_id)
                        self.repo.add(obj)
                    session.commit()
            elif entity_type == "projects":
                project_id = kwargs["id"]
                existing_project = self.repo.list(Project, {"id": project_id})
                if not existing_project:
                    raise ValueError(
                        f"Project with id {project_id} is not found!")
                tag = self.repo.list(BaseTag, {"text": kwargs["text"]})
                if not tag:
                    tag = BaseTag(**kwargs)
                    self.repo.add(tag)
                    session.commit()
                    tag_id = tag.id
                else:
                    tag_id = tag[0].id

                existing_project_tag = self.repo.list(ProjectTag, {
                    "project": project_id,
                    "tag": tag_id
                })
                if not existing_project_tag:
                    project_ids = get_ids(project_id)
                    for project_id in project_ids:
                        obj = ProjectTag(id=project_id, tag_id=tag_id)
                        self.repo.add(obj)
                    session.commit()
        elif command == "comment":
            if entity_type == "tasks":
                existing_task = self.repo.list(Task, {"id": kwargs["id"]})
                if not existing_task:
                    raise ValueError(
                        f"Task with id {kwargs['id']} is not found!")
                obj = TaskCommentary(**kwargs)
                self.repo.add(obj)
                self.repo.update(Task, kwargs["id"], {"modification_date": datetime.now()})
                session.commit()
            elif entity_type == "projects":
                existing_project = self.repo.list(Project,
                                                  {"id": kwargs["id"]})
                if not existing_project:
                    raise ValueError(
                        f"Project with id {kwargs['id']} is not found!")
                obj = ProjectCommentary(**kwargs)
                self.repo.add(obj)
                session.commit()
            elif entity_type == "epics":
                epic = self.repo.list(Epic, {"id": kwargs["id"]})
                if not epic:
                    raise ValueError(
                        f"Epic with id {kwargs['id']} is not found!")
                obj = EpicCommentary(**kwargs)
                self.repo.add(obj)
                session.commit()
            elif entity_type == "stories":
                epic = self.repo.list(Story, {"id": kwargs["id"]})
                if not epic:
                    raise ValueError(
                        f"Story with id {kwargs['id']} is not found!")
                obj = StoryCommentary(**kwargs)
                self.repo.add(obj)
                session.commit()
            elif entity_type == "sprints":
                sprint = self.repo.list(Sprint, {"id": kwargs["id"]})
                if not sprint:
                    raise ValueError(
                        f"Sprint with id {kwargs['id']} is not found!")
                obj = SprintCommentary(**kwargs)
                self.repo.add(obj)
                session.commit()
        elif command == "delete":
            if entity not in (Task, ):
                raise ValueError("'delete' operation only allowed for tasks!")
            task_ids = get_ids(kwargs.get("id"))
            for task_id in task_ids:
                if not (task := self.repo.get_by_id(Task, task_id)):
                    exit(f"Task id {task_id} is not found")
                else:
                    if all(key == "id" for key in kwargs.keys()):
                        deletion_reason = input(
                            f"Provide brief explanation why you're deleting task <{task_id}>: "
                        )
                        if deletion_reason:
                            self.execute("comment", "tasks", {
                                "id": task_id,
                                "text": deletion_reason
                            })
                        kwargs.update({"status": "DELETED"})
                        self.execute("update", entity_type, kwargs)
                if epic_id := kwargs.get("epic_id"):
                    epic = self.repo.list(Epic, {"id": epic_id})
                    if not epic:
                        exit(f"Epic id {epic_id} is not found")
                    if not (epic_task := self.repo.list(
                            EpicTask, {
                                "task": task_id,
                                "epic": epic_id
                            })):
                        exit("task is not in epic")
                    self.repo.delete(EpicTask, epic_task[0].id)
                if story_id := kwargs.get("story_id"):
                    story = self.repo.list(Story, {"id": story_id})
                    if not story:
                        exit(f"Story id {story_id} is not found")
                    if not (story_task := self.repo.list(
                            StoryTask, {
                                "task": task_id,
                                "story": story_id
                            })):
                        exit("task is not in story")
                    self.repo.delete(StoryTask, story_task[0].id)
                if sprint_id := kwargs.get("sprint_id"):
                    sprint = self.repo.list(Sprint, {"id": sprint_id})
                    if not sprint:
                        exit(f"Sprint id {sprint_id} is not found")
                    if sprint[0].status.name == "COMPLETED":
                        exit("Cannot add task to a finished sprint")
                    if not (sprint_task := self.repo.list(
                            SprintTask, {
                                "task": task_id,
                                "sprint": sprint_id
                            })):
                        exit("task is not in sprint")
                    self.repo.delete(SprintTask, sprint_task[0].id)
            session.commit()
        elif command == "connect":
            if entity not in (Task, Project):
                raise ValueError("'connect' operation only allowed for tasks and project!")
            task_ids = get_ids(kwargs.get("id"))
            for task_id in task_ids:
                if entity_type == "projects":
                    asana_project_id = kwargs.get("external_project")
                    asana_project = self.repo.get_by_id(AsanaProject, task_id)
                    if not asana_project:
                        obj = AsanaProject(id=task_id, asana_project_id=asana_project_id)
                        self.repo.add(obj)
                    else:
                        self.repo.update(AsanaProject, task_id, {"asana_project_id": asana_project_id})
                elif entity_type == "tasks":
                    asana_task_id = kwargs.get("external_task")
                    asana_task = self.repo.get_by_id(AsanaProject, task_id)
                    if not asana_task:
                        obj = AsanaTask(project=task.project, id=task_id, asana_task_id=kwargs.get("external_task"))
                        self.repo.add(obj)
                    else:
                        self.repo.update(AsanaTask, task_id, {"asana_task_id": asana_task_id})
            self.repo.session.commit()
        elif command == "add":
            if entity not in (Task, ):
                raise ValueError("'add' operation only allowed for tasks!")
            task_ids = get_ids(kwargs.get("id"))
            for task_id in task_ids:
                if not (added_task := self.repo.get_by_id(Task, task_id)):
                    exit(f"Task id {task_id} is not found")

                if epic_id := kwargs.get("epic_id"):
                    epic = self.repo.list(Epic, {"id": epic_id})
                    if not epic:
                        exit(f"Epic id {epic_id} is not found")
                    obj = EpicTask(task=task_id, epic=epic_id)
                    if self.repo.list(EpicTask, {
                            "task": obj.task,
                            "epic": obj.epic
                    }):
                        exit("task already added to epic")
                    self.repo.add(obj)
                if story_id := kwargs.get("story_id"):
                    story = self.repo.list(Story, {"id": story_id})
                    if not story:
                        exit(f"Story id {story_id} is not found")
                    obj = StoryTask(task=task_id, story=story_id)
                    if self.repo.list(StoryTask, {
                            "task": obj.task,
                            "story": obj.story
                    }):
                        exit("task already added to story")
                    self.repo.add(obj)
                if sprint_id := kwargs.get("sprint_id"):
                    sprint = self.repo.get_by_id(Sprint, sprint_id)
                    if not sprint:
                        exit(f"Sprint id {sprint_id} is not found")
                    if sprint.status.name == "COMPLETED":
                        exit("Cannot add task to a finished sprint")
                    obj = SprintTask(task=task_id,
                                     sprint=sprint_id,
                                     is_active_link=True)
                    if self.repo.list(SprintTask, {
                            "task": obj.task,
                            "sprint": obj.sprint
                    }):
                        exit("task already added to sprint")
                    self.repo.add(obj)
                    sprint_task_id = obj.id
                else:
                    sprint_task_id = None
                # Update task status and due date
                if sprint_task_id:
                    task_params = {"id": added_task.id}
                    if added_task.status.name == "BACKLOG":
                        task_params.update({"status": "TODO"})
                    if not added_task.due_date or added_task.due_date > sprint.end_date:
                        task_params.update({"due_date": sprint.end_date})
                    self.execute("update", "tasks", task_params)
                if story_points := kwargs.get("story_points"):
                    if not sprint_task_id:
                        sprint_task = self.execute("get", "sprint_tasks",
                                                   {"task": task_id})
                        if not sprint_task:
                            exit(
                                f"Task id {task_id} is not part of any sprint")
                        else:
                            sprint_task_id = sprint_task[0].id
                    self.repo.update(SprintTask, sprint_task_id,
                                     {"story_points": story_points})
            session.commit()
        elif command == "create":
            kwargs["created_by"] = services.lookup_user_id(
                self.config.get("user"), self.repo)
            obj = entity(**kwargs)
            if entity in (Task, Project):
                if (existing_obj := self.repo.list(entity,
                                                   {"name": obj.name})):
                    print("Found existing entity\n")
                    existing_id = str(existing_obj[0].id)
                    self.execute("show", entity_type, {"id": existing_id})
                    answer = input(
                        "Do you want to create entity anyway? [Y/n] ")
                    if answer.lower() != "y":
                        exit()
            self.repo.add(obj)
            session.commit()
            if tag_text := kwargs.get("tags"):
                self.execute("tag", entity_type, {
                    "id": obj.id,
                    "text": tag_text
                })
            if collaborator_name := kwargs.get("collaborators"):
                self.execute("collaborate", entity_type, {
                    "id": obj.id,
                    "name": collaborator_name
                })
            if hasattr(obj, "project") and obj.project:
                project = services.lookup_project_name(obj.project, self.repo)
            else:
                project = None
            self.printer.print_new_object(obj, project)
            logger.info("<create> %s: %s", entity_type, obj.id)
            return entity, obj
        elif command == "update":
            if (task_id := kwargs.get("id")):
                del kwargs["id"]
                tasks = get_ids(task_id)
                for task in tasks:
                    if (old_values := self.repo.list(entity, {"id": task})):
                        old_values = old_values[0]
                        old_settings = {
                            key: getattr(old_values, key)
                            for key in kwargs.keys()
                        }
                        if old_settings != kwargs:
                            new_kwargs = {}
                            for k, v in kwargs.items():
                                if old_settings[k] != v:
                                    new_kwargs[k] = v
                            self.repo.update(entity, task, new_kwargs)
                            logger.info("<update> %s: %s", entity_type, task)
                            now = datetime.now()
                            if entity_type in ("tasks", "projects"):
                                for key, value in new_kwargs.items():
                                    try:
                                        old_value = old_settings[key].name
                                    except:
                                        old_value = old_settings[key]
                                    if entity_type == "projects":
                                        event = ProjectEvent
                                    elif entity_type == "tasks":
                                        event = TaskEvent
                                    self.repo.add(
                                        event(task, key, old_value, value,
                                              now))
                            if hasattr(entity, "modification_date"):
                                self.repo.update(entity, task, {"modification_date": now})
                        else:
                            print(
                                "No changes were proposed to the existing entity"
                            )
                    session.commit()
                return entity, None
            else:
                raise ValueError("No task_id is provided")
        elif command == "edit":
            if (task_id := kwargs.get("id")):
                tasks = get_ids(task_id)
                for task in tasks:
                    if task.isdigit():
                        entities = self.repo.get_by_id(entity, task)
                        task_id = task
                    else:
                        entities = self.repo.get(entity, task)
                        task_id = entities.id
                    if entity_type in ("tasks", "projects", "sprints", "epics", "stories"):
                        task = entities
                        if kwargs.get("description"):
                            current_entry = task.description
                            current_type = "description"
                        elif kwargs.get("name"):
                            current_entry = task.name
                            current_type = "name"
                        elif kwargs.get("goal"):
                            current_entry = task.goal
                            current_type = "goal"
                        else:
                            raise ValueError(
                                "Either name or description should be specified!"
                            )

                    message_template = generate_message_template(
                        CurrentEntry(current_entry, current_type, entity_type), task,
                        self.repo)
                    with tempfile.NamedTemporaryFile(suffix=".tmp") as tf:
                        tf.write(message_template.encode("utf-8"))
                        tf.flush()
                        run(["vim", "+2", tf.name])
                        tf.seek(0)
                        new_entry = tf.read()

                    updated_entry = []
                    new_entry = new_entry.decode("utf-8").rstrip()
                    for i, row in enumerate(new_entry.split("\n")):
                        if not row.startswith("#") and row:
                            if row.startswith("--"):
                                break
                            updated_entry.append(row)
                new_kwargs = {
                    "id": task_id,
                    current_type: " ".join(updated_entry)
                }
                self.execute("update", entity_type, new_kwargs)
                return entities, None, None
        elif command == "start":
            sprint_id = kwargs.get("id")
            kwargs.update({"status": "ACTIVE"})
            self.execute("update", "sprints", kwargs)
            [sprint] = self.execute("get", "sprints", {"id": sprint_id})
            for sprint_task in sprint.tasks:
                task = sprint_task.tasks
                task_params = {"id": task.id}
                if task.status.name == "BACKLOG":
                    task_params.update({"status": "TODO"})
                if not task.due_date or task.due_date > sprint.end_date:
                    task_params.update({"due_date": sprint.end_date})
                self.execute("update", "tasks", task_params)
        elif command == "show":
            if len(kwargs) == 1 and "project" in kwargs:
                kwargs["id"] = kwargs["project"]
                del kwargs["project"]
            print_options = printer.PrintOptions(
                show_history=bool(kwargs.pop("show_history", False)),
                show_commentaries=bool(kwargs.pop("show_commentaries", False)),
                show_completed=bool(kwargs.pop("show_completed", False)))
            if kwargs.pop("partial_project_view", False):
                print_options.show_epics = bool(kwargs.pop("epics", False))
                print_options.show_tasks = bool(kwargs.pop("tasks", False))
                print_options.show_stories = bool(kwargs.pop("stories", False))
            if entity_type == "tasks":
                print_options.show_completed = True
            if not (task_id := kwargs.get("id")):
                if entity_type == "sprints":
                    active_sprint = self.repo.list(Sprint,
                                                   {"status": "ACTIVE"})
                    if len(active_sprint) == 1:
                        task_id = active_sprint[0].id
                    else:
                        exit(
                            "More than 1 active sprint, please specify the sprint_id"
                        )
            tasks = get_ids(task_id)
            if "id" in kwargs:
                kwargs.pop("id")
            for task in tasks:
                if task.isdigit():
                    entities = self.repo.list(entity, {"id": task})
                else:
                    entities = self.repo.list(entity, {"name": task})
                self.printer.print_entity(task, entity_type, entities,
                                          self.repo, print_options, kwargs)
                logger.info("<show> %s: %s", entity_type, task_id)
            return entities, None, None
        elif command == "done":
            if entity_type not in ("tasks", "sprints", "stories", "epics"):
                raise ValueError("can complete only tasks and sprints")
            elif entity_type == "tasks":
                kwargs.update({"status": "DONE"})
                self.execute("update", entity_type, kwargs)
                self.console.print(
                    "[green]Yay! You've just completed a task![/green]")
            elif entity_type in ("sprints", "epics", "stories"):
                kwargs.update({"status": "COMPLETED"})
                [entity] = self.execute("get", entity_type,
                                        {"id": kwargs.get("id")})

                if entity_type == "sprints":
                    entity.complete(entity.tasks)
                else:
                    entity.complete(entity.tasks)
                    for entity_task in entity.tasks:
                        self.execute("update", "tasks", {"id": entity_task.task, "status": "DONE"})
                self.execute("update", entity_type, kwargs)
        elif command == "track":
            if entity_type != "tasks":
                raise ValueError("can track only tasks")
            if "hours" in kwargs and "minutes" in kwargs:
                exit("specify only -H (hours) or -M (minutes) value")
            if "hours" in kwargs or "minutes" in kwargs:
                if "hours" in kwargs:
                    kwargs["time_spent_minutes"] = float(kwargs["hours"]) * 60
                    del kwargs["hours"]
                else:
                    kwargs["time_spent_minutes"] = float(kwargs["minutes"])
                    del kwargs["minutes"]
                kwargs["task"] = kwargs["id"]
                del kwargs["id"]
                self.execute("create", "time_entry", kwargs)
            else:
                exit("tracking missing -H (hours) or -M (minutes) value")
        else:
            raise ValueError(f"Uknown command: {command}")

    def _read_config(self) -> Dict[str, Any]:
        with open(f"{self.home_dir}/.terka/config.yaml", "r",
                  encoding="utf-8") as f:
            config = yaml.safe_load(f)
        return config


def get_ids(ids: Union[str, int]) -> List[Union[int, str]]:
    id_string = str(ids)
    if "," in id_string:
        ids = id_string.split(",")
    elif ".." in id_string:
        task_range = id_string.split("..")
        ids = range(int(task_range[0]), int(task_range[1]) + 1)
    else:
        ids = [id_string]
    return ids



from typing import Dict, List, Optional


def time_spent(session, start_date: str,
               end_date: str) -> List[Dict[str, str]]:
    results = session.execute(
        """
    SELECT
        STRFTIME("%Y-%m-%d", creation_date) AS date,
        ROUND(SUM(time_spent_minutes) / 60, 2) AS time_spent_hours
    FROM time_tracker_entries
    WHERE
        creation_date >= :start_date
        AND creation_date <= :end_date
    GROUP BY 1
    """, dict(start_date=start_date, end_date=end_date))
    return [dict(r) for r in results]


def sprint_task_ids(session,
                    sprint_id: Optional[int] = None) -> List[Dict[int, int]]:
    results = session.execute(
        """
    SELECT
        task, sprint
    FROM sprint_tasks
    WHERE
        sprint = :sprint_id
    """, dict(sprint_id=sprint_id))
    return [dict(r) for r in results]


def external_connectors_asana_projects(session) -> List[Dict[int, str]]:
    results = session.execute(
        """
    SELECT
        id, asana_project_id
    FROM 'external_connectors.asana.projects'
    """)
    return [dict(r) for r in results]


def external_connectors_asana_tasks(session, project_id: str) -> List[Dict[int, str]]:
    results = session.execute(
        """
    SELECT
        id, asana_task_id
    FROM 'external_connectors.asana.tasks'
    WHERE project = :project_id
    """, dict(project_id=project_id))
    return [dict(r) for r in results]



from typing import Any, Dict, List, Optional, Tuple
from dataclasses import dataclass
from operator import attrgetter
import inspect
from copy import deepcopy

from collections import defaultdict
from datetime import datetime, date, timedelta
import plotext as plt
import rich
from rich.console import Console
from rich.table import Table
from statistics import mean, median

from terka.service_layer import services, views
from terka.service_layer.ui import TerkaTask


@dataclass
class PrintOptions:
    show_tasks: bool = True
    show_history: bool = False
    show_commentaries: bool = False
    show_completed: bool = False
    show_epics: bool = True
    show_stories: bool = True


class Printer:

    def __init__(self, repo, box=rich.box.SIMPLE) -> None:
        self.console = Console()
        self.box = box
        self.repo = repo

    def print_new_object(self, obj, project):
        table = Table(box=self.box)
        attributes = self._get_attributes(obj)
        for column, value in attributes:
            if value:
                table.add_column(column)
        table.add_row(*list(zip(*attributes))[1])
        if table.row_count:
            self.console.print(table)

    def print_history(self, entities):
        table = Table(box=self.box)
        print("History:")
        for column in ("date", "type", "old_value", "new_value"):
            table.add_column(column)
        for event in entities:
            table.add_row(event.date.strftime("%Y-%m-%d %H:%M"),
                          event.type.name, event.old_value, event.new_value)
        if table.row_count:
            self.console.print(table)

    def print_commentaries(self, entities):
        table = Table(box=self.box)
        print("Comments:")
        for column in ("date", "text"):
            table.add_column(column)
        entities.sort(key=lambda c: c.date, reverse=False)
        for event in entities:
            table.add_row(event.date.strftime("%Y-%m-%d %H:%M"), event.text)
        if table.row_count:
            self.console.print(table)

    def print_entity(self,
                     task,
                     entity_type,
                     entities,
                     repo,
                     print_options: PrintOptions = PrintOptions(),
                     kwargs: Optional[Dict[str, Any]] = None):
        if entity_type == "sprints":
            if entities:
                self.print_sprint(entities, repo, print_options, kwargs)
            else:
                exit(f"No sprint with id '{task}' found!")
        if entity_type in ("epics", "stories"):
            if entities:
                self.print_composite(entities, repo, print_options, entity_type)
            else:
                exit(f"No {entity_type} with id '{task}' found!")
        if entity_type == "projects":
            if entities:
                self.print_project(entities, print_options, kwargs)
            else:
                exit(f"No project '{task}' found!")
        if entity_type == "tasks":
            if entities:
                self.print_task(entities, repo, print_options)
            else:
                print(f"No task with id '{task}' found!")

    def print_entities(self, entities, type, repo, custom_sort, print_options):
        if type == "projects":
            entities.sort(key=self._sort_open_tasks, reverse=True)
            self.print_project(entities, print_options)
        elif type == "tasks":
            if custom_sort:
                entities.sort(key=lambda c: getattr(c, custom_sort),
                              reverse=False)
            else:
                entities.sort(key=lambda c:
                              (c.status.value, c.priority.value
                               if hasattr(c.priority, "value") else 0),
                              reverse=True)
            self.print_task(entities=entities,
                            repo=repo,
                            print_options=print_options,
                            custom_sort=custom_sort)
        elif type == "tags":
            self.print_tag(entities=entities)
        elif type == "users":
            self.print_user(entities=entities)
        elif type == "sprints":
            self.print_sprint(entities=entities,
                              repo=repo,
                              print_options=print_options)
        elif type in ("epics", "stories"):
            self.print_composite(entities=entities,
                            repo=repo,
                            print_options=print_options,
                            composite_type=type)
        else:
            self.print_default_entity(self, entities)

    def print_user(self, entities):
        table = Table(box=self.box)
        for column in ("id", "name"):
            table.add_column(column)
        seen_users = set()
        for entity in entities:
            if entity.name not in seen_users:
                table.add_row(f"[red]{entity.id}[/red]", entity.name)
                seen_users.add(entity.name)
        if table.row_count:
            self.console.print(table)

    def print_tag(self, entities):
        table = Table(box=self.box)
        for column in ("id", "text"):
            table.add_column(column)
        seen_tags = set()
        for entity in entities:
            if entity.text not in seen_tags:
                table.add_row(f"[red]{entity.id}[/red]", entity.text)
                seen_tags.add(entity.text)
        if table.row_count:
            self.console.print(table)

    def print_default_entity(self, entities):
        table = Table(box=self.box)
        for column in ("id", "date", "text"):
            table.add_column(column)
        for entity in entities:
            table.add_row(f"[red]{entity.id}[/red]",
                          entity.date.strftime("%Y-%m-%d %H:%M"), entity.text)
        if table.row_count:
            self.console.print(table)

    def print_composite(self, entities, repo, print_options, composite_type):
        table = Table(box=self.box, title=composite_type.upper(), expand=True)
        for column in ("id", "name", "description", "project", "tasks"):
            table.add_column(column, style="bold")
        for i, entity in enumerate(entities):
            tasks = []
            completed_tasks = []
            for entity_task in entity.tasks:
                task = entity_task.tasks
                if task.status.name not in ("DONE", "DELETED"):
                    tasks.append(task)
                else:
                    completed_tasks.append(task)
            try:
                project_obj = services.lookup_project_name(
                    entity.project, repo)
                project = project_obj.name
            except:
                project = None
            if i == 0 or tasks or print_options.show_completed:
                table.add_row(f"E{entity.id}", str(entity.name),
                              entity.description, project, str(len(tasks)))
        if table.row_count:
            self.console.print(table)
        if print_options.show_tasks and tasks:
            self.print_task(entities=tasks,
                            repo=repo,
                            print_options=print_options,
                            show_window=False)
        if print_options.show_tasks and completed_tasks:
            self.print_task(entities=completed_tasks,
                            repo=repo,
                            print_options=print_options,
                            show_window=False)
        if i == 0 and print_options.show_commentaries and (
                commentaries := entity.commentaries):
            self.print_commentaries(commentaries)

    def print_sprint(self, entities, repo, print_options, kwargs=None):
        table = Table(box=rich.box.SQUARE_DOUBLE_HEAD)
        for column in ("id", "start_date", "end_date", "goal", "status",
                       "open tasks", "tasks", "velocity", "collaborators",
                       "time_spent"):
            table.add_column(column)
        for i, entity in enumerate(entities):
            story_points = []
            tasks = []
            collaborators = []
            collaborators_dict = defaultdict(int)
            for sprint_task in entity.tasks:
                story_points.append(sprint_task.story_points)
                task = sprint_task.tasks
                tasks.append(task)
                if task_collaborators := task.collaborators:
                    for collaborator in task.collaborators:
                        collaborators_dict[collaborator.users.
                                           name] += sprint_task.story_points
                else:
                    collaborators_dict["Unknown"] += sprint_task.story_points

            for user, story_point in sorted(collaborators_dict.items(),
                                            key=lambda x: x[1],
                                            reverse=True):
                collaborators.append(f"{user} ({round(story_point, 2)})")

            collaborators_string = ", ".join(collaborators)
            open_tasks = [
                task for task in tasks
                if task.status.name not in ("DONE", "DELETED")
            ]
            time_spent_sum = sum([
                entry.time_spent_minutes for task in tasks
                for entry in task.time_spent
            ])

            time_spent = self._format_time_spent(time_spent_sum)
            table.add_row(str(entity.id), str(entity.start_date),
                          str(entity.end_date), entity.goal,
                          entity.status.name, str(len(open_tasks)),
                          str(len(tasks)), str(round(sum(story_points),
                                                     2)), collaborators_string,
                          str(time_spent))
        if table.row_count:
            self.console.print(table)

        if print_options.show_tasks:
            task_print_options = deepcopy(print_options)
            task_print_options.show_commentaries = False
            self.print_task(entities=tasks,
                            repo=repo,
                            print_options=task_print_options,
                            story_points=story_points,
                            kwargs=kwargs,
                            show_window=False)
        if i == 0 and print_options.show_commentaries and (
                commentaries := entities[0].commentaries):
            self.print_commentaries(commentaries)
        time_entries = views.time_spent(repo.session, entity.start_date,
                                        entity.end_date)
        dates = [entry.get("date") for entry in time_entries]
        times = [entry.get("time_spent_hours") for entry in time_entries]
        plt.date_form('Y-m-d')
        plt.plot_size(100, 15)
        plt.title("Time tracker")
        plt.bar(dates, times)
        plt.show()

    def print_project(self, entities, print_options, kwargs=None):
        table = Table(box=rich.box.SQUARE_DOUBLE_HEAD, expand=True)
        non_active_projects = Table(box=rich.box.SQUARE_DOUBLE_HEAD)
        for column in ("id", "name", "description", "status", "open_tasks",
                       "overdue", "stale", "backlog", "todo", "in_progress",
                       "review", "done", "median_task_age"):
            table.add_column(column)
        for column in ("id", "name", "description", "status", "open_tasks"):
            non_active_projects.add_column(column)
        for entity in entities:
            # if entity.status.name != "ACTIVE":
            #     continue
            open_tasks = self._sort_open_tasks(entity)
            if open_tasks > 0 and entity.status.name == "ACTIVE":
                overdue_tasks = [
                    task for task in entity.tasks
                    if task.due_date and task.due_date <= datetime.now().date(
                    ) and task.status.name not in ("DELETED", "DONE")
                ]
                stale_tasks = []
                for task in entity.tasks:
                    if (event_history :=
                            task.history) and task.status.name in (
                                "TODO", "IN_PROGRESS", "REVIEW"):
                        for event in event_history:
                            if max([
                                    event.date for event in event_history
                            ]) < (datetime.today() - timedelta(days=5)):
                                stale_tasks.append(task)
                stale_tasks = list(set(stale_tasks))
                median_task_age = round(
                    median([(datetime.now() - task.creation_date).days
                            for task in entity.tasks
                            if task.status.name not in ("DELETED", "DONE")]))
                backlog = self._count_task_status(entity.tasks, "BACKLOG")
                todo = self._count_task_status(entity.tasks, "TODO")
                in_progress = self._count_task_status(entity.tasks,
                                                      "IN_PROGRESS")
                review = self._count_task_status(entity.tasks, "REVIEW")
                done = self._count_task_status(entity.tasks, "DONE")

                if overdue_tasks:
                    entity_id = f"[red]{entity.id}[/red]"
                else:
                    entity_id = f"[green]{entity.id}[/green]"
                table.add_row(f"{entity_id}", entity.name, entity.description,
                              entity.status.name, str(open_tasks),
                              str(len(overdue_tasks)), str(len(stale_tasks)),
                              str(backlog), str(todo), str(in_progress),
                              str(review), str(done), str(median_task_age))
            else:
                non_active_projects.add_row(str(entity.id), entity.name,
                                            entity.description,
                                            entity.status.name,
                                            str(open_tasks))

        if table.row_count:
            self.console.print(table)
        if non_active_projects.row_count:
            self.console.print("[green]Inactive projects[/green]")
            self.console.print(non_active_projects)
        non_task_view_options = deepcopy(print_options)
        non_task_view_options.show_tasks = False
        if print_options.show_epics and (epics := entity.epics):
            self.console.print("")
            self.print_composite(epics, self.repo, non_task_view_options, "epics")
        if print_options.show_stories and (stories := entity.stories):
            self.console.print("")
            self.print_composite(stories, self.repo, non_task_view_options, "stories")
        if print_options.show_tasks and (tasks := entity.tasks):
            task_print_options = deepcopy(print_options)
            task_print_options.show_commentaries = False
            self.console.print("")
            self.print_task(entities=tasks,
                            repo=self.repo,
                            print_options=task_print_options,
                            show_window=False,
                            kwargs=kwargs)
        if print_options.show_commentaries and (commentaries :=
                                                entity.commentaries):
            self.print_commentaries(commentaries)
        if print_options.show_history and (history := entity.history):
            self.print_history(history)

    def print_task(self,
                   entities,
                   repo,
                   print_options,
                   custom_sort=None,
                   show_window=True,
                   story_points=None,
                   kwargs=None):
        table = Table(box=self.box, title="TASKS")
        if story_points:
            default_columns = ("id", "name", "description", "story points",
                               "status", "priority", "project", "due_date",
                               "tags", "collaborators", "time_spent")
        else:
            default_columns = ("id", "name", "description", "status",
                               "priority", "project", "due_date", "tags",
                               "collaborators", "time_spent")
        #TODO: Add printing for only a single task
        # if (entities[0].status.name == "DONE"):
        #     console.print(f"[green]task is completed on [/green]")
        # else:
        #     active_in_days = datetime.now() - entities[0].creation_date
        #     console.print(f"[blue]task is active {active_in_days.days} days[/blue]")
        entities = list(entities)
        if kwargs:
            entities = self._get_filtered_entities(entities, kwargs)
        if not story_points:
            if custom_sort:
                entities.sort(key=lambda c: getattr(c, custom_sort),
                              reverse=False)
            else:
                entities.sort(key=lambda c: (c.status.value, c.priority.value),
                              reverse=True)
        table, completed_tasks, completed_story_points = self._print_task(
            table=table,
            entities=entities,
            default_columns=default_columns,
            repo=repo,
            story_points=story_points,
            show_window=show_window)
        if table.row_count:
            self.console.print(table)
        if print_options.show_completed and completed_tasks:
            table = Table(box=self.box)
            table, *rest = self._print_task(
                table=table,
                entities=completed_tasks,
                default_columns=default_columns,
                repo=repo,
                story_points=completed_story_points,
                show_window=show_window,
                all_tasks=False)
            if table.row_count:
                self.console.print(f"[green]****COMPLETED TASKS*****[/green]")
                self.console.print(table)

    def _get_attributes(self, obj) -> List[Tuple[str, str]]:
        import inspect
        attributes = []
        for name, value in inspect.getmembers(obj):
            if not name.startswith("_") and not inspect.ismethod(value):
                if not value:
                    continue
                elif name in ("created_by", "assignee"):
                    attributes.append(
                        (name, services.lookup_user_name(value, self.repo)))
                elif name == "project":
                    project = services.lookup_project_name(value, self.repo)
                    attributes.append((name, project.name))
                elif hasattr(value, "name"):
                    attributes.append((name, value.name))
                elif isinstance(value, datetime):
                    attributes.append((name, value.strftime("%Y-%m-%d %H:%M")))
                elif isinstance(value, set):
                    values = value.pop()
                    attributes.append((name, str(values)))
                else:
                    attributes.append((name, str(value)))
        return attributes

    def _sort_open_tasks(self, entities):
        return len([
            task for task in entities.tasks
            if task.status.name not in ("DONE", "DELETED")
        ])

    def _count_task_status(self, tasks, status: str) -> int:
        return len([task for task in tasks if task.status.name == status])

    def _calculate_time_spent(self, entity) -> str:
        time_spent_sum = sum(
            [entry.time_spent_minutes for entry in entity.time_spent])
        return self._format_time_spent(time_spent_sum)

    def _format_time_spent(self, time_spent: int) -> str:
        time_spent_hours = time_spent // 60
        time_spent_minutes = time_spent % 60
        if time_spent_hours and time_spent_minutes:
            time_spent = f"{time_spent_hours}H:{time_spent_minutes}M"
        elif time_spent_hours:
            time_spent = f"{time_spent_hours}H"
        elif time_spent_minutes:
            time_spent = f"{time_spent_minutes}M"
        else:
            time_spent = ""
        return time_spent

    def _get_filtered_entities(self, entities, kwargs):
        if kwargs:
            filtering_attributes = set(list(kwargs))
            temp_entities = []
            attributes = [
                name for name, value in inspect.getmembers(entities[0])
                if not name.startswith("_") and not inspect.ismethod(value)
            ]
            filtering_attributes = filtering_attributes.intersection(
                set(attributes))
            attribute_getter = attrgetter(*filtering_attributes)
            filtered_entities = []
            for entity in entities:
                should_add = False
                for key, value in kwargs.items():
                    if key in filtering_attributes:
                        attribute_value = getattr(entity, key)
                        if hasattr(attribute_value, "name"):
                            if attribute_value.name != value:
                                should_add = False
                                break
                            else:
                                should_add = True
                                continue
                        elif attribute_value != value:
                            should_add = False
                            break
                        else:
                            should_add = True
                            continue
                if should_add:
                    filtered_entities.append(entity)
            entities = list(filtered_entities)
        return entities

    def _print_task(self,
                    table,
                    entities,
                    default_columns,
                    repo,
                    story_points=None,
                    all_tasks=True,
                    show_window=True):
        if all_tasks:
            completed_tasks = []
        else:
            completed_tasks = None
            completed_story_points = None
        if story_points:
            completed_story_points = []
            entities = [(entity, story_point) for entity, story_point in
                        sorted(zip(entities, story_points),
                               key=lambda x: x[0].status.value,
                               reverse=True)]
        else:
            completed_story_points = None
        for column in default_columns:
            table.add_column(column)
        for entity in entities:
            if story_points:
                story_point = entity[1]
                entity = entity[0]
            else:
                story_point = None
            if tags := entity.tags:
                tag_texts = sorted([tag.base_tag.text for tag in list(tags)])
                tag_string = ",".join(tag_texts)
            else:
                tag_string = ""
            if collaborators := entity.collaborators:
                collaborators_texts = sorted([
                    collaborator.users.name
                    for collaborator in list(collaborators)
                    if collaborator.users
                ])
                collaborator_string = ",".join(collaborators_texts)
            else:
                collaborator_string = ""
            if comments := entity.commentaries:
                comments.sort(key=lambda c: c.date, reverse=False)
            if history := entity.history:
                history.sort(key=lambda c: c.date, reverse=False)
            try:
                project_obj = services.lookup_project_name(
                    entity.project, repo)
                project = project_obj.name
            except:
                project = None
            priority = entity.priority.name if hasattr(entity.priority,
                                                       "name") else "UNKNOWN"
            time_spent = self._calculate_time_spent(entity)
            if entity.status.name in ("DELETED", "DONE") and all_tasks:
                completed_tasks.append(entity)
                if story_points:
                    completed_story_points.append(story_point)
                continue
            entity_name = f"{entity.name}" if not comments else f"{entity.name} [blue][{len(comments)}][/blue]"
            if not all_tasks:
                entity_id = str(entity.id)
            elif entity.due_date and entity.due_date <= date.today():
                entity_id = f"[red]{entity.id}[/red]"
            elif (event_history := entity.history) and entity.status.name in (
                    "TODO", "IN_PROGRESS", "REVIEW"):
                if max([event.date for event in event_history
                        ]) < (datetime.today() - timedelta(days=5)):
                    entity_id = f"[yellow]{entity.id}[/yellow]"
                else:
                    entity_id = str(entity.id)
            else:
                entity_id = str(entity.id)
            if story_points:
                table.add_row(entity_id, entity_name, entity.description,
                              str(story_point), entity.status.name, priority,
                              project, str(entity.due_date), tag_string,
                              collaborator_string, str(time_spent))
            else:
                table.add_row(entity_id, entity_name, entity.description,
                              entity.status.name, priority, project,
                              str(entity.due_date), tag_string,
                              collaborator_string, str(time_spent))
        if table.row_count == 1 and show_window:
            app = TerkaTask(entity=entity,
                            project=project,
                            history=history,
                            commentaries=comments)
            app.run()
        # # TODO: not working
        # if print_options.show_commentaries and (commentaries :=
        #                                         entity.commentaries):
        #     self.print_commentaries(commentaries)
        # # TODO: not working
        # if print_options.show_history and (history := entity.history):
        #     self.print_history(history)
        return table, completed_tasks, completed_story_points



def save_updates(entity_type, task, old_settings, new_kwargs):
    now = datetime.now()
    for key, value in new_kwargs.items():
        try:
            old_value = old_settings[key].name
        except:
            old_value = old_settings[key]
        self.repo.add(
            Event(entity_type, task, now, key,
                  old_value, value))




from typing import Set
from enum import Enum
import logging
from datetime import datetime
from dataclasses import dataclass

from .task import Task

logger = logging.getLogger(__name__)


class SprintStatus(Enum):
    PLANNED = 1
    ACTIVE = 2
    COMPLETED = 3


class Sprint:

    def __init__(self,
                 start_date: datetime = None,
                 end_date: datetime = None,
                 status: str = "PLANNED",
                 goal: str = None,
                 **kwargs) -> None:
        if start_date.date()  < datetime.today().date():
            raise ValueError(f"start date cannot be less than today")
        if not start_date:
            raise ValueError(
                "Please provide start date for the sprint in YYYY-MM-DD format"
            )
        self.start_date = start_date
        if not end_date:
            raise ValueError(
                "Please provide end date for the sprint in YYYY-MM-DD format")
        self.end_date = end_date
        if end_date.date() < start_date.date():
            raise ValueError(f"Sprint end date cannot be less than start date")
        if end_date.date() < datetime.today().date():
            raise ValueError(f"Sprint end date cannot be less than today")
        self.status = self._validate_status(status)
        self.goal = goal
        self.tasks: Set[Task, ...] = set()
        self.is_completed = False

    def _validate_status(self, status):
        if status and status not in [s.name for s in SprintStatus]:
            raise ValueError(f"{status} is invalid status")
        else:
            return status

    def complete(self, tasks) -> None:
        incompleted_tasks = list()
        for task in tasks:
            if task.tasks.status.name != "DONE":
                incompleted_tasks.append(task)
        if incompleted_tasks:
            logging.warning("[Sprint %d]: %d tasks haven't been completed",
                            self.id, len(incompleted_tasks))
        self.is_completed = True


@dataclass
class SprintTask:
    sprint: int
    task: int
    story_points: int = 0
    #TODO: add actual_time_spent: int = 0
    is_active_link: bool = True



from datetime import datetime


class TimeTrackerEntry:

    def __init__(self,
                 task: int,
                 time_spent_minutes: float,
                 creation_date: datetime = datetime.now(),
                 **kwargs
                 ):
        if not isinstance(creation_date, datetime):
            raise ValueError(
                "creation_date should be of type datetime.datetime!")
        self.task = task
        self.time_spent_minutes = time_spent_minutes
        self.creation_date = creation_date



from typing import Any, Dict, Optional
import os
import yaml

from terka.adapters.repository import AbsRepository
from terka.domain.project import Project
from terka.domain.user import User


def update_config(update_pair: Dict[str, Any]):
    config = get_config()
    config.update(update_pair)
    with open(f"{home_dir}/.terka/config.yaml", "w") as f:
        yaml.dump(config, f)

def get_config():
    home_dir = os.path.expanduser('~')
    with open(f"{home_dir}/.terka/config.yaml",
              "r",
              encoding="utf-8") as f:
        config = yaml.safe_load(f)
    return config


def lookup_project_id(project_name: str, repo: AbsRepository) -> int:
    projects = project_name.split(",")
    if project_name.startswith("NOT"):
        projects = [
            f"NOT:{project}"
            for project in project_name.replace("NOT:", "").split(",")
        ]
        negate = True
    else:
        projects = project_name.split(",")
        negate = False
    returned_projects = []
    returned_projects_set = set()
    for project_name in projects:
        if not (project := repo.list(Project, project_name)):
            print(
                f"Creating new project: {project_name}. Do you want to continue (Y/n)?"
            )
            answer = input()
            while answer.lower() != "y":
                print("Provide a project name: ")
                project_name = input()
                print(
                    f"Creating new project: {project_name}. Do you want to continue (Y/n)?"
                )
                answer = input()
            project = Project(project_name)
            repo.add(project)
            repo.session.commit()
        if isinstance(project, list):
            project_ids = [project.id for project in project]
            if negate:
                if returned_projects_set:
                    returned_projects_set =returned_projects_set.intersection(set(project_ids))
                else:
                     returned_projects_set = returned_projects_set.union(set(project_ids))

            else:
                returned_projects.extend(project_ids)
        else:
            returned_projects.append(project.id)
    if returned_projects_set:
        return list(returned_projects_set)
    if len(returned_projects) == 1:
        return returned_projects[0]
    return returned_projects


def lookup_user_id(user_name: str, repo: AbsRepository) -> Optional[int]:
    user = repo.list(User, {"name": user_name})
    if user:
        return user[0].id
    return None


def lookup_user_name(user_id: str, repo: AbsRepository) -> Optional[str]:
    user = repo.list(User, {"id": user_id})
    if user:
        return user[0].name
    return None


def lookup_project_name(project_id: int, repo: AbsRepository) -> Optional[str]:
    project = repo.list(Project, {"id": project_id})
    if project:
        return project[0]
    return None



import pathlib
from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent
# README = (HERE.parent / "README.md").read_text()

setup(
    name="terka",
    version="1.7.0",
    description="CLI utility for creating and managing tasks in a terminal",
    # long_description=README,
    long_description_content_type="text/markdown",
    author="Andrei Markin",
    author_email="andrey.markin.ppc@gmail.com",
    license="Apache 2.0",
    url="https://github.com/AndreyMarkinPPC/terka",
    packages=find_packages(),
    include_package_data=True,
    package_data={"": ["service_layer/*.css"]},
    install_requires=[
        "sqlalchemy", "pyaml", "rich", "textual==0.5.0", "plotext==5.2.8"
    ],
    setup_requires=["pytest-runner"],
    tests_requires=["pytest"],
    entry_points={
        "console_scripts": [
            "terka=terka.entrypoints.cli:main"
        ]
    }
)



class BaseTag:

    def __init__(self, text: str, **kwargs: str) -> None:
        self.text = text

    def __repr__(self):
        return f"<Tag> {self.text}"


class TagMixin:
    def __repr__(self):
        return f"{self.base_tag.text}"


class TaskTag(TagMixin):

    def __init__(self,
                 id: int,
                 tag_id: int,
                 **kwargs: str) -> None:
        self.task = id
        self.tag= tag_id


class ProjectTag(TagMixin):

    def __init__(self,
                 id: int,
                 tag_id: int,
                 **kwargs: str) -> None:
        self.project = id
        self.tag= tag_id
